@@ -69,9 +69,10 @@ fi
 
 CRYPTOPP=cryptopp
 CRYPTOPP_VERSION=820
+CRYPTOPP_VERSION2=$(echo "$CRYPTOPP_VERSION" | sed 's/\(.\)/_\1/g')  # convert version "820" to "_8_2_0"
 CRYPTOPP_SOURCE_FILE=cryptopp${CRYPTOPP_VERSION}.zip
 CRYPTOPP_SOURCE_FOLDER=${CRYPTOPP}/${CRYPTOPP}
-CRYPTOPP_DOWNLOAD_URL=http://www.cryptopp.com/${CRYPTOPP_SOURCE_FILE}
+CRYPTOPP_DOWNLOAD_URL=https://github.com/weidai11/cryptopp/releases/download/CRYPTOPP${CRYPTOPP_VERSION2}/${CRYPTOPP_SOURCE_FILE}
 CRYPTOPP_SHA1="b042d2f0c93410abdec7c12bcd92787d019f8da1"
 
 SQLITE=sqlite
@@ -85,17 +86,18 @@ SQLITE_SHA1="350fa5ccedc70f4979d7f954fba9525542809ba2"
 
 CURL=curl
 CURL_VERSION=8.1.1
-C_ARES_VERSION=1.19.1
 CURL_EXTRA="--disable-dict --disable-file --disable-ftp --disable-gopher --disable-imap --disable-ldap --disable-ldaps --disable-mime --disable-netrc --disable-pop3 --disable-proxy --disable-rtsp --disable-smb --disable-smtp --disable-telnet --disable-tftp --disable-manual"
 CURL_SOURCE_FILE=curl-${CURL_VERSION}.tar.gz
 CURL_SOURCE_FOLDER=curl-${CURL_VERSION}
 CURL_DOWNLOAD_URL=http://curl.haxx.se/download/${CURL_SOURCE_FILE}
 CURL_SHA1="5ff2ecaa4a68ecc06434644ce76d9837e99e7d1d"
 
+C_ARES_VERSION=1.19.1
+C_ARES_VERSION2=1_19_1
 ARES_SOURCE_FILE=c-ares-${C_ARES_VERSION}.tar.gz
 ARES_SOURCE_FOLDER=c-ares-${C_ARES_VERSION}
 ARES_CONFIGURED=${CURL}/${ARES_SOURCE_FOLDER}/Makefile.inc
-ARES_DOWNLOAD_URL=http://c-ares.haxx.se/download/${ARES_SOURCE_FILE}
+ARES_DOWNLOAD_URL=https://github.com/c-ares/c-ares/releases/download/cares-${C_ARES_VERSION2}/${ARES_SOURCE_FILE}
 ARES_SHA1="99566278e4ed4b261891aa62c8b88227bf1a2823"
 
 CRASHLYTICS=crashlytics
@@ -439,7 +441,7 @@ if [ ! -f ${LIBUV}/${LIBUV_SOURCE_FILE}.ready ]; then
 
         pushd ${LIBUV}/${LIBUV} &>> ${LOG_FILE}
         ./autogen.sh &>> ${LOG_FILE}
-        ./configure --host "${TARGET_HOST}" --with-pic --disable-shared --prefix="${BASE_PATH}/${LIBUV}/${LIBUV}"/libuv-android-${ABI} &>> ${LOG_FILE}
+        LDFLAGS+="-Wl,-z,max-page-size=16384" ./configure --host "${TARGET_HOST}" --with-pic --disable-shared --prefix="${BASE_PATH}/${LIBUV}/${LIBUV}"/libuv-android-${ABI} &>> ${LOG_FILE}
         make clean &>> ${LOG_FILE}
         make -j${JOBS} &>> ${LOG_FILE}
         make install &>> ${LOG_FILE}
@@ -576,7 +578,7 @@ if [ ! -f ${CURL}/${CURL_SOURCE_FILE}.ready ]; then
 	    fi
 
         pushd ${CURL}/ares &>> ${LOG_FILE}
-        ./configure --host "${TARGET_HOST}" --with-pic --disable-shared --prefix="${BASE_PATH}/${CURL}"/ares/ares-android-${ABI} &>> ${LOG_FILE}
+        LDFLAGS+="-Wl,-z,max-page-size=16384" ./configure --host "${TARGET_HOST}" --with-pic --disable-shared --prefix="${BASE_PATH}/${CURL}"/ares/ares-android-${ABI} &>> ${LOG_FILE}
         make clean &>> ${LOG_FILE}
         make -j${JOBS} &>> ${LOG_FILE}
         make install &>> ${LOG_FILE}
@@ -584,7 +586,7 @@ if [ ! -f ${CURL}/${CURL_SOURCE_FILE}.ready ]; then
 
         pushd ${CURL}/${CURL} &>> ${LOG_FILE}
 
-        LIBS=-lc++ ./configure --host "${TARGET_HOST}" --with-pic --disable-shared --prefix="${BASE_PATH}/${CURL}/${CURL}"/curl-android-${ABI} --with-ssl="${BASE_PATH}/${OPENSSL}/${OPENSSL}/${SSL_SUFFIX}" \
+        LDFLAGS+="-Wl,-z,max-page-size=16384" LIBS=-lc++ ./configure --host "${TARGET_HOST}" --with-pic --disable-shared --prefix="${BASE_PATH}/${CURL}/${CURL}"/curl-android-${ABI} --with-ssl="${BASE_PATH}/${OPENSSL}/${OPENSSL}/${SSL_SUFFIX}" \
         --enable-ares="${BASE_PATH}/${CURL}"/ares/ares-android-${ABI} ${CURL_EXTRA} &>> ${LOG_FILE}
         make clean &>> ${LOG_FILE}
         make -j${JOBS} &>> ${LOG_FILE}
@@ -612,7 +614,7 @@ if [ ! -f ${ICU}/${ICU_SOURCE_FILE}.ready ]; then
 
     mkdir -p linux && cd linux
 
-    CONFIGURE_LINUX_OPTIONS="--enable-static --enable-shared=no --enable-extras=no --enable-strict=no --enable-icuio=no --enable-layout=no --enable-layoutex=no --enable-tools=yes --enable-tests=no --enable-samples=no --enable-dyload=no"
+    LDFLAGS+="-Wl,-z,max-page-size=16384" CONFIGURE_LINUX_OPTIONS="--enable-static --enable-shared=no --enable-extras=no --enable-strict=no --enable-icuio=no --enable-layout=no --enable-layoutex=no --enable-tools=yes --enable-tests=no --enable-samples=no --enable-dyload=no"
     ../source/runConfigureICU Linux CFLAGS="-Os" CXXFLAGS="--std=c++11" ${CONFIGURE_LINUX_OPTIONS} &>> ${LOG_FILE}
 
     make -j${JOBS} &>> ${LOG_FILE}
@@ -648,7 +650,7 @@ if [ ! -f ${ICU}/${ICU_SOURCE_FILE}.ready ]; then
         export PATH=$ANDROID_TOOLCHAIN/bin:$PATH
 
         rm -rf ${ANDROID_TOOLCHAIN} &>> ${LOG_FILE}
-        $NDK_ROOT/build/tools/make-standalone-toolchain.sh --arch=${ARCH} --platform=${APP_PLATFORM} --install-dir=${ANDROID_TOOLCHAIN} &>> ${LOG_FILE}
+        $NDK_ROOT/build/tools/make_standalone_toolchain.py --arch=${ARCH} --api=${API_LEVEL} --install-dir=${ANDROID_TOOLCHAIN} &>> ${LOG_FILE}
 
         CONFIGURE_ANDROID_OPTIONS="--host=${HOST} --enable-static --enable-shared=no --enable-extras=no --enable-strict=no --enable-icuio=no --enable-layout=no --enable-layoutex=no --enable-tools=no --enable-tests=no --enable-samples=no --enable-dyload=no -with-cross-build=$CROSS_BUILD_DIR"
 
