@@ -24,7 +24,9 @@
 #define NODEMANAGER_H 1
 
 #include <map>
+#include <limits>
 #include <set>
+#include <vector>
 #include "node.h"
 #include "types.h"
 
@@ -39,46 +41,219 @@ class NodeSerialized;
 class NodeSearchFilter
 {
 public:
-    NodeSearchFilter(ShareType_t st = NO_SHARES) : mShareType(st) {}
+    static constexpr char TAG_DELIMITER = ',';
 
-    template<class T>
-    void copyFrom(const T& f, ShareType_t shareType = NO_SHARES)
+    enum class BoolFilter
     {
-        mNameFilter = f.byName() ? f.byName() : std::string(); // get it as const char*
-        mNodeType = static_cast<nodetype_t>(f.byNodeType()); // get it as int
-        mMimeCategory = static_cast<MimeType_t>(f.byCategory()); // get it as int
-        mExcludeSensitive = f.bySensitivity();
-        mLocationHandle = f.byLocationHandle();
-        mShareType = shareType;
-        mCreationLowerLimit = f.byCreationTimeLowerLimit();
-        mCreationUpperLimit = f.byCreationTimeUpperLimit();
+        disabled = 0,
+        onlyTrue,
+        onlyFalse,
+    };
+
+    enum class TextQueryJoiner
+    {
+        logicalAnd,
+        logicalOr,
+    };
+
+    void byAncestors(std::vector<handle>&& ancs) { assert(ancs.size() == 3); mLocationHandles.swap(ancs); }
+    void setIncludedShares(ShareType_t s) { mIncludedShares = s; }
+
+    void byName(const std::string& name)
+    {
+        mNameFilter = name;
     }
 
-    const std::string& byName() const { return mNameFilter; }
+    void byNodeType(nodetype_t nodeType)
+    {
+        assert(nodeType >= nodetype_t::TYPE_UNKNOWN && nodeType <= nodetype_t::FOLDERNODE);
+        mNodeType = nodeType;
+    }
+
+    void byCategory(const MimeType_t category)
+    {
+        mMimeCategory = category;
+    }
+
+    void bySensitivity(BoolFilter boolFilter)
+    {
+        mExcludeSensitive = boolFilter;
+    }
+
+    void byFavourite(const BoolFilter byFav)
+    {
+        mFavouriteFilterOption = byFav;
+    }
+
+    void byLocationHandle(const handle location)
+    {
+        mLocationHandles = {location, UNDEF, UNDEF};
+    }
+
+    void byCreationTimeLowerLimitInSecs(int64_t creationLowerLimit)
+    {
+        mCreationLowerLimit = creationLowerLimit;
+    }
+
+    void byCreationTimeUpperLimitInSecs(int64_t creationUpperLimit)
+    {
+        mCreationUpperLimit = creationUpperLimit;
+    }
+
+    void byModificationTimeLowerLimitInSecs(int64_t modificationLowerLimit)
+    {
+        mModificationLowerLimit = modificationLowerLimit;
+    }
+
+    void byModificationTimeUpperLimitInSecs(int64_t modificationUpperLimit)
+    {
+        mModificationUpperLimit = modificationUpperLimit;
+    }
+
+    void byDescription(const std::string& description)
+    {
+        mDescriptionFilter = escapeWildCards(description);
+    }
+
+    void byTag(const std::string& tag)
+    {
+        mTagFilter = escapeWildCards(tag);
+        mTagFilterContainsSeparator = mTagFilter.getText().find(TAG_DELIMITER) != std::string::npos;
+    }
+
+    void useAndForTextQuery(const bool useAnd)
+    {
+        mUseAndForTextQuery = useAnd;
+    }
+
+    const std::string& byName() const
+    {
+        return mNameFilter.getText();
+    }
     nodetype_t byNodeType() const { return mNodeType; }
     MimeType_t byCategory() const { return mMimeCategory; }
-    bool bySensitivity() const { return mExcludeSensitive; }
+    BoolFilter byFavourite() const { return mFavouriteFilterOption; }
+    BoolFilter bySensitivity() const { return mExcludeSensitive; }
 
-    // recursive look-ups (searchNodes): represents 'ancestor';
-    // non-recursive look-ups (getChildren): represents 'parent'.
-    handle byLocationHandle() const { return mLocationHandle; }
+    // recursive look-ups (searchNodes)
+    const std::vector<handle>& byAncestorHandles() const { return mLocationHandles; }
+    // non-recursive look-ups (getChildren)
+    handle byParentHandle() const { assert(!mLocationHandles.empty()); return mLocationHandles[0]; }
 
-    // recursive look-ups (searchNodes): type of share where search is restricted to;
+    // recursive look-ups (searchNodes): type of shares to be included when searching;
     // non-recursive look-ups (getChildren): ignored.
-    ShareType_t byShareType() const { return mShareType; }
+    ShareType_t includedShares() const { return mIncludedShares; }
 
     int64_t byCreationTimeLowerLimit() const { return mCreationLowerLimit; }
     int64_t byCreationTimeUpperLimit() const { return mCreationUpperLimit; }
 
+    int64_t byModificationTimeLowerLimit() const { return mModificationLowerLimit; }
+
+    int64_t byModificationTimeUpperLimit() const
+    {
+        return mModificationUpperLimit;
+    }
+
+    const std::string& byDescription() const
+    {
+        return mDescriptionFilter.getText();
+    }
+
+    const std::string& byTag() const
+    {
+        return mTagFilter.getText();
+    }
+
+    bool useAndForTextQuery() const
+    {
+        return mUseAndForTextQuery;
+    }
+
+    bool hasNodeType() const
+    {
+        return mNodeType != TYPE_UNKNOWN;
+    }
+
+    bool hasCreationTimeLimits() const
+    {
+        return mCreationLowerLimit || mCreationUpperLimit;
+    }
+
+    bool hasModificationTimeLimits() const
+    {
+        return mModificationLowerLimit || mModificationUpperLimit;
+    }
+
+    bool hasCategory() const
+    {
+        return mMimeCategory != MIME_TYPE_UNKNOWN;
+    }
+
+    bool hasName() const
+    {
+        return !mNameFilter.getText().empty();
+    }
+
+    bool hasDescription() const
+    {
+        return !mDescriptionFilter.getText().empty();
+    }
+
+    bool hasTag() const
+    {
+        return !mTagFilter.getText().empty();
+    }
+
+    bool hasFav() const
+    {
+        return mFavouriteFilterOption != BoolFilter::disabled;
+    }
+
+    bool hasSensitive() const
+    {
+        return mExcludeSensitive != BoolFilter::disabled;
+    }
+
+    bool isValidNodeType(const nodetype_t nodeType) const;
+    bool isValidCreationTime(const int64_t time) const;
+    bool isValidModificationTime(const int64_t time) const;
+    bool isValidCategory(const MimeType_t category, const nodetype_t nodeType) const;
+    bool isValidName(const uint8_t* testName) const;
+    bool isValidDescription(const uint8_t* testDescription) const;
+    bool isValidTagSequence(const uint8_t* tagSequence) const;
+    bool isValidFav(const bool isNodeFav) const;
+    bool isValidSensitivity(const bool isNodeSensitive) const;
+
 private:
-    std::string mNameFilter;
+    TextPattern mNameFilter;
     nodetype_t mNodeType = TYPE_UNKNOWN;
     MimeType_t mMimeCategory = MIME_TYPE_UNKNOWN;
-    bool mExcludeSensitive = false;
-    handle mLocationHandle = UNDEF;
-    ShareType_t mShareType = NO_SHARES;
+    BoolFilter mFavouriteFilterOption = BoolFilter::disabled;
+    BoolFilter mExcludeSensitive = BoolFilter::disabled;
+    std::vector<handle> mLocationHandles {UNDEF, UNDEF, UNDEF}; // always contain 3 items
+    ShareType_t mIncludedShares = NO_SHARES;
     int64_t mCreationLowerLimit = 0;
     int64_t mCreationUpperLimit = 0;
+    int64_t mModificationLowerLimit = 0;
+    int64_t mModificationUpperLimit = 0;
+    TextPattern mDescriptionFilter;
+    TextPattern mTagFilter;
+    bool mTagFilterContainsSeparator{false};
+    bool mUseAndForTextQuery{true};
+
+    static bool isDocType(const MimeType_t t);
+};
+
+class NodeSearchPage
+{
+public:
+    NodeSearchPage(size_t startingOffset, size_t size) : mOffset(startingOffset), mSize(size) {}
+    const size_t& startingOffset() const { return mOffset; }
+    const size_t& size() const { return mSize; }
+
+private:
+    size_t mOffset;
+    size_t mSize;
 };
 
 /**
@@ -104,68 +279,47 @@ public:
     void reset();
 
     // Take node ownership
-    typedef map<NodeHandle,  set<Node*>> MissingParentNodes;
-    bool addNode(Node* node, bool notify, bool isFetching, MissingParentNodes& missingParentNodes);
+    typedef map<NodeHandle,  set<std::shared_ptr<Node>>> MissingParentNodes;
+    bool addNode(std::shared_ptr<Node> node, bool notify, bool isFetching, MissingParentNodes& missingParentNodes);
     bool updateNode(Node* node);
     // removeNode() --> it's done through notifypurge()
 
     // if node is not available in memory, it's loaded from DB
-    Node *getNodeByHandle(NodeHandle handle);
+    std::shared_ptr<Node> getNodeByHandle(NodeHandle handle);
 
     // read children from DB and load them in memory
-    node_list getChildren(const Node *parent, CancelToken cancelToken = CancelToken());
+    sharedNode_list getChildren(const Node *parent, CancelToken cancelToken = CancelToken());
 
-    node_vector getChildren(const NodeSearchFilter& filter, CancelToken cancelFlag);
-
-    // read children from type (folder or file) from DB and load them in memory
-    node_vector getChildrenFromType(const NodeHandle &parent, nodetype_t type, CancelToken cancelToken);
+    sharedNode_vector getChildren(const NodeSearchFilter& filter, int order, CancelToken cancelFlag, const NodeSearchPage& page);
 
     // get up to "maxcount" nodes, not older than "since", ordered by creation time
     // Note: nodes are read from DB and loaded in memory
-    node_vector getRecentNodes(unsigned maxcount, m_time_t since);
+    sharedNode_vector getRecentNodes(unsigned maxcount,
+                                     m_time_t since,
+                                     bool excludeSensitives = false);
 
-    // Search nodes containing 'searchString' in its name
-    // Returned nodes are children of 'nodeHandle' (at any level)
-    // If 'nodeHandle' is UNDEF, search includes the whole account
-    // If a cancelFlag is passed, it must be kept alive until this method returns
-    /** @deprecated Use searchNodes(const NodeSearchFilter...) instead */
-    node_vector search(NodeHandle ancestorHandle, const char* searchString, bool recursive, Node::Flags requiredFlags, Node::Flags excludeFlags, Node::Flags excludeRecursiveFlags, CancelToken cancelFlag);
+    sharedNode_vector searchNodes(const NodeSearchFilter& filter, int order, CancelToken cancelFlag, const NodeSearchPage& page);
 
-    node_vector searchNodes(const NodeSearchFilter& filter, CancelToken cancelFlag);
+    std::set<std::string> getAllNodeTags(const char* searchString, CancelToken cancelFlag);
 
-    /** @deprecated Use searchNodes(const NodeSearchFilter...) instead */
-    node_vector getInSharesWithName(const char *searchString, CancelToken cancelFlag);
-    /** @deprecated Use searchNodes(const NodeSearchFilter...) instead */
-    node_vector getOutSharesWithName(const char *searchString, CancelToken cancelFlag);
-    /** @deprecated Use searchNodes(const NodeSearchFilter...) instead */
-    node_vector getPublicLinksWithName(const char *searchString, CancelToken cancelFlag);
-
-
-    node_vector getNodesByFingerprint(FileFingerprint& fingerprint);
-    node_vector getNodesByOrigFingerprint(const std::string& fingerprint, Node *parent);
-    Node *getNodeByFingerprint(FileFingerprint &fingerprint);
+    sharedNode_vector getNodesByFingerprint(FileFingerprint& fingerprint);
+    sharedNode_vector getNodesByOrigFingerprint(const std::string& fingerprint, Node *parent);
+    std::shared_ptr<Node> getNodeByFingerprint(FileFingerprint &fingerprint);
 
     // Return a first level child node whose name matches with 'name'
     // Valid values for nodeType: FILENODE, FOLDERNODE
     // Note: if not found among children loaded in RAM (and not all children are loaded), it will search in DB
     // Hint: ensure all children are loaded if this method is called for all children of a folder
-    Node* childNodeByNameType(const Node *parent, const std::string& name, nodetype_t nodeType);
+    std::shared_ptr<Node> childNodeByNameType(const Node *parent, const std::string& name, nodetype_t nodeType);
 
     // Returns ROOTNODE, INCOMINGNODE, RUBBISHNODE (In case of logged into folder link returns only ROOTNODE)
     // Load from DB if it's necessary
-    node_vector getRootNodes();
+    sharedNode_vector getRootNodes();
 
-    /** @deprecated Use searchNodes(const NodeSearchFilter...) instead */
-    node_vector getNodesWithInShares(); // both, top-level and nested ones
-    /** @deprecated Use searchNodes(const NodeSearchFilter...) instead */
-    node_vector getNodesWithOutShares();
-
-    node_vector getNodesWithPendingOutShares();
-
-    /** @deprecated Use searchNodes(const NodeSearchFilter...) instead */
-    node_vector getNodesWithLinks();
-    /** @deprecated Use searchNodes(const NodeSearchFilter...) instead */
-    node_vector getNodesByMimeType(MimeType_t mimeType, NodeHandle ancestorHandle, Node::Flags requiredFlags, Node::Flags excludeFlags, Node::Flags excludeRecursiveFlags, CancelToken cancelFlag);
+    sharedNode_vector getNodesWithInShares(); // both, top-level and nested ones
+    sharedNode_vector getNodesWithOutShares();
+    sharedNode_vector getNodesWithPendingOutShares();
+    sharedNode_vector getNodesWithLinks();
 
     std::vector<NodeHandle> getFavouritesNodeHandles(NodeHandle node, uint32_t count);
     size_t getNumberOfChildrenFromNode(NodeHandle parentHandle);
@@ -185,16 +339,16 @@ public:
 
     // Use blob received as parameter to generate a node
     // Used to generate nodes from old cache
-    Node* getNodeFromBlob(const string* nodeSerialized);
+    std::shared_ptr<Node> getNodeFromBlob(const string* nodeSerialized);
 
     // attempt to apply received keys to decrypt node's keys
     void applyKeys(uint32_t appliedKeys);
 
     // add node to the notification queue
-    void notifyNode(Node* node, node_vector* nodesToReport = nullptr);
+    void notifyNode(std::shared_ptr<Node> node, sharedNode_vector* nodesToReport = nullptr);
 
     // for consistently notifying when updating node counters
-    void setNodeCounter(Node* n, const NodeCounter &counter, bool notify, node_vector* nodesToReport);
+    void setNodeCounter(std::shared_ptr<Node> n, const NodeCounter &counter, bool notify, sharedNode_vector* nodesToReport);
 
     // process notified/changed nodes from 'mNodeNotify': dump changes to DB
     void notifyPurge();
@@ -215,20 +369,21 @@ public:
     NodeCounter getCounterOfRootNodes();
 
     // update the counter of 'n' when its parent is updated (from 'oldParent' to 'n.parent')
-    void updateCounter(Node &n, Node *oldParent);
+    void updateCounter(std::shared_ptr<Node> n, std::shared_ptr<Node> oldParent);
 
     // true if 'h' is a rootnode: cloud, inbox or rubbish bin
     bool isRootNode(NodeHandle h) const;
 
     // Set values to mClient.rootnodes for ROOTNODE, INBOX and RUBBISH
-    bool setrootnode(Node* node);
+    bool setrootnode(std::shared_ptr<Node> node);
 
     // Add fingerprint to mFingerprint. If node isn't going to keep in RAM
     // node isn't added
     FingerprintPosition insertFingerprint(Node* node);
     // Remove fingerprint from mFingerprint
-    void removeFingerprint(Node* node);
+    void removeFingerprint(Node* node, bool unloadNode = false);
     FingerprintPosition invalidFingerprintPos();
+    std::list<std::shared_ptr<Node>>::const_iterator invalidCacheLRUPos() const;
 
     // Node has received last updates and it's ready to store in DB
     void saveNodeInDb(Node *node);
@@ -261,10 +416,33 @@ public:
     // Initialize node counters and create indexes at DB
     void initCompleted();
 
+    std::shared_ptr<Node> getNodeFromNodeManagerNode(NodeManagerNode& nodeManagerNode);
+
+    void insertNodeCacheLRU(std::shared_ptr<Node> node);
+
+    void increaseNumNodesInRam();
+    void decreaseNumNodesInRam();
+
+    uint64_t getCacheLRUMaxSize() const;
+    void setCacheLRUMaxSize(uint64_t cacheLRUMaxSize);
+
+    uint64_t getNumNodesAtCacheLRU() const;
+
     // true when the filesystem has been initialized
+    // i.e., when nodes have been fully loaded from a fetchnodes or from cache
     bool ready();
 
 private:
+    class NoKeyLogger
+    {
+    public:
+        void log(const Node& node) const;
+
+    private:
+        // How many no key nodes has been counted for logging
+        mutable std::atomic_int mCount{1};
+    };
+
     MegaClient& mClient;
 
 #if defined(DEBUG)
@@ -280,16 +458,24 @@ private:
     // interface to handle accesses to "nodes" table
     DBTableNodes* mTable = nullptr;
 
+    // logger with rate limitting for no key
+    static NoKeyLogger mNoKeyLogger;
+
     // root nodes (files, vault, rubbish)
     struct Rootnodes
     {
         NodeHandle files;
         NodeHandle vault;
         NodeHandle rubbish;
+        std::map<nodetype_t, std::shared_ptr<Node> > mRootNodes;
 
+        // minimum expected number of root nodes (min num of root nodes may vary depending on
+        // client type i.e password manager)
+        static constexpr uint8_t MIN_NUM_ROOT_NODES{3};
         // returns true if the 'h' provided matches any of the rootnodes.
         // (when logged into folder links, the handle of the folder is set to 'files')
         bool isRootNode(NodeHandle h) const { return (h == files || h == vault || h == rubbish); }
+        void clear();
     } rootnodes;
 
     class FingerprintContainer : public fingerprint_set
@@ -297,6 +483,7 @@ private:
     public:
         bool allFingerprintsAreLoaded(const FileFingerprint *fingerprint) const;
         void setAllFingerprintLoaded(const FileFingerprint *fingerprint);
+        void removeAllFingerprintLoaded(const FileFingerprint *fingerprint);
         void clear();
 
     private:
@@ -308,16 +495,18 @@ private:
     // Stores nodes that have been loaded in RAM from DB (not necessarily all of them)
     std::map<NodeHandle, NodeManagerNode> mNodes;
 
-    uint64_t mNodesInRam = 0;
+    uint64_t mCacheLRUMaxSize = std::numeric_limits<uint64_t>::max();
+    std::list<std::shared_ptr<Node> > mCacheLRU;
+
+    std::atomic<uint64_t> mNodesInRam;
 
     // nodes that have changed and are pending to notify to app and dump to DB
-    node_vector mNodeNotify;
+    sharedNode_vector mNodeNotify;
 
-    Node* getNodeInRAM(NodeHandle handle);
-    void saveNodeInRAM(Node* node, bool isRootnode, MissingParentNodes& missingParentNodes);    // takes ownership
+    shared_ptr<Node> getNodeInRAM(NodeHandle handle);
+    void saveNodeInRAM(std::shared_ptr<Node> node, bool isRootnode, MissingParentNodes& missingParentNodes);    // takes ownership
 
-    /** @deprecated */
-    node_vector getNodesWithSharesOrLink_internal(ShareType_t shareType);
+    sharedNode_vector getNodesWithSharesOrLink_internal(ShareType_t shareType);
 
     enum OperationType
     {
@@ -327,39 +516,41 @@ private:
 
     // Update a node counter for 'origin' and its subtree (recursively)
     // If operationType is INCREASE, nc is added, in other case is decreased (ie. upon deletion)
-    void updateTreeCounter(Node* origin, NodeCounter nc, OperationType operation, node_vector* nodesToReport);
+    void updateTreeCounter(std::shared_ptr<Node> origin, NodeCounter nc, OperationType operation, sharedNode_vector* nodesToReport);
 
     // returns nullptr if there are unserialization errors. Also triggers a full reload (fetchnodes)
-    Node* getNodeFromNodeSerialized(const NodeSerialized& nodeSerialized);
+    shared_ptr<Node> getNodeFromNodeSerialized(const NodeSerialized& nodeSerialized);
 
     // reads from DB and loads the node in memory
-    Node* unserializeNode(const string*, bool fromOldCache);
+    shared_ptr<Node> unserializeNode(const string*, bool fromOldCache);
 
     // returns the counter for the specified node, calculating it recursively and accessing to DB if it's neccesary
-    NodeCounter calculateNodeCounter(const NodeHandle &nodehandle, nodetype_t parentType, Node *node, bool isInRubbish);
+    NodeCounter calculateNodeCounter(const NodeHandle &nodehandle, nodetype_t parentType, std::shared_ptr<Node> node, bool isInRubbish);
 
     // Container storing FileFingerprint* (Node* in practice) ordered by fingerprint
     FingerprintContainer mFingerPrints;
 
     // Return a node from Data base, node shouldn't be in RAM previously
-    Node* getNodeFromDataBase(NodeHandle handle);
+    shared_ptr<Node> getNodeFromDataBase(NodeHandle handle);
 
     // Returns root nodes without nested in-shares
-    node_vector getRootNodesAndInshares();
+    sharedNode_vector getRootNodesAndInshares();
 
     // Process unserialized nodes read from DB
     // Avoid loading nodes whose ancestor is not ancestorHandle. If ancestorHandle is undef load all nodes
     // If a valid cancelFlag is passed and takes true value, this method returns without complete operation
     // If a valid object is passed, it must be kept alive until this method returns.
-    node_vector processUnserializedNodes(const std::vector<std::pair<NodeHandle, NodeSerialized>>& nodesFromTable, NodeHandle ancestorHandle = NodeHandle(), CancelToken cancelFlag = CancelToken());
+    sharedNode_vector processUnserializedNodes(const std::vector<std::pair<NodeHandle, NodeSerialized>>& nodesFromTable, NodeHandle ancestorHandle = NodeHandle(), CancelToken cancelFlag = CancelToken());
 
-    node_vector searchNodes_internal(const NodeSearchFilter& filter, CancelToken cancelFlag);
-    node_vector processUnserializedNodes(const std::vector<std::pair<NodeHandle, NodeSerialized>>& nodesFromTable, const NodeSearchFilter& filter, CancelToken cancelFlag = CancelToken());
-    node_vector getChildren_internal(const NodeSearchFilter& filter, CancelToken cancelFlag);
-    node_vector processUnserializedChildren(const std::vector<std::pair<NodeHandle, NodeSerialized>>& childrenFromTable, const NodeSearchFilter& filter, CancelToken cancelFlag = CancelToken());
+    sharedNode_vector searchNodes_internal(const NodeSearchFilter& filter, int order, CancelToken cancelFlag, const NodeSearchPage& page);
+    sharedNode_vector processUnserializedNodes(const std::vector<std::pair<NodeHandle, NodeSerialized>>& nodesFromTable, CancelToken cancelFlag);
+    sharedNode_vector getChildren_internal(const NodeSearchFilter& filter, int order, CancelToken cancelFlag, const NodeSearchPage& page);
+    sharedNode_vector getRecentNodes_internal(const NodeSearchPage& page, m_time_t since);
+
+    std::set<std::string> getAllNodeTags_internal(const char* searchString, CancelToken cancelFlag);
 
     // node temporary in memory, which will be removed upon write to DB
-    unique_ptr<Node> mNodeToWriteInDb;
+    std::shared_ptr<Node> mNodeToWriteInDb;
 
     // Stores (or updates) the node in the DB. It also tries to decrypt it for the last time before storing it.
     void putNodeInDb(Node* node) const;
@@ -367,6 +558,9 @@ private:
     // true when the NodeManager has been inicialized and contains a valid filesystem
     bool mInitialized = false;
 
+    // flag that determines if null root nodes error has already been reported
+    bool mNullRootNodesReported{false};
+
     // These are all the "internal" versions of the public interfaces.
     // This is to avoid confusion where public functions used to call other public functions
     // but that introudces confusion about where the mutex gets locked.
@@ -378,32 +572,25 @@ private:
     // It's quite a verbose approach, but at least simple, easy to understand, and easy to get right.
     void setTable_internal(DBTableNodes *table);
     void reset_internal();
-    bool addNode_internal(Node* node, bool notify, bool isFetching, MissingParentNodes& missingParentNodes);
+    bool addNode_internal(std::shared_ptr<Node> node, bool notify, bool isFetching, MissingParentNodes& missingParentNodes);
     bool updateNode_internal(Node* node);
-    Node *getNodeByHandle_internal(NodeHandle handle);
-    node_list getChildren_internal(const Node *parent, CancelToken cancelToken = CancelToken());
-    node_vector getChildrenFromType_internal(const NodeHandle& parent, nodetype_t type, CancelToken cancelToken);
-    node_vector getRecentNodes_internal(unsigned maxcount, m_time_t since);
 
-    /** @deprecated */
-    node_vector search_internal(NodeHandle ancestorHandle, const char* searchString, bool recursive, Node::Flags requiredFlags, Node::Flags excludeFlags, Node::Flags excludeRecursiveFlags, CancelToken cancelFlag);
-    /** @deprecated */
-    node_vector getInSharesWithName_internal(const char *searchString, CancelToken cancelFlag);
-    /** @deprecated */
-    node_vector getOutSharesWithName_internal(const char *searchString, CancelToken cancelFlag);
-    /** @deprecated */
-    node_vector getPublicLinksWithName_internal(const char *searchString, CancelToken cancelFlag);
+    /**
+     * @brief Manages null root nodes error server event (just once in NodeManager lifetime)
+     * This method sends an event to stats server and prints a log error to inform about this
+     * scenario.
+     */
+    void reportNullRootNodes(const size_t rootNodesSize);
 
-    node_vector getNodesByFingerprint_internal(FileFingerprint& fingerprint);
-    node_vector getNodesByOrigFingerprint_internal(const std::string& fingerprint, Node *parent);
-    Node *getNodeByFingerprint_internal(FileFingerprint &fingerprint);
-    Node* childNodeByNameType_internal(const Node *parent, const std::string& name, nodetype_t nodeType);
-    node_vector getRootNodes_internal();
+    std::shared_ptr<Node> getNodeByHandle_internal(NodeHandle handle);
+    sharedNode_list getChildren_internal(const Node* parent,
+                                         CancelToken cancelToken = CancelToken());
 
-    /** @deprecated */
-    node_vector getNodesWithInShares_internal(); // both, top-level and nested ones
-    /** @deprecated */
-    node_vector getNodesByMimeType_internal(MimeType_t mimeType, NodeHandle ancestorHandle, Node::Flags requiredFlags, Node::Flags excludeFlags, Node::Flags excludeRecursiveFlags, CancelToken cancelFlag);
+    sharedNode_vector getNodesByFingerprint_internal(FileFingerprint& fingerprint);
+    sharedNode_vector getNodesByOrigFingerprint_internal(const std::string& fingerprint, Node *parent);
+    std::shared_ptr<Node> getNodeByFingerprint_internal(FileFingerprint &fingerprint);
+    std::shared_ptr<Node> childNodeByNameType_internal(const Node *parent, const std::string& name, nodetype_t nodeType);
+    sharedNode_vector getRootNodes_internal();
 
     std::vector<NodeHandle> getFavouritesNodeHandles_internal(NodeHandle node, uint32_t count);
     size_t getNumberOfChildrenFromNode_internal(NodeHandle parentHandle);
@@ -411,16 +598,16 @@ private:
     bool isAncestor_internal(NodeHandle nodehandle, NodeHandle ancestor, CancelToken cancelFlag);
     void removeChanges_internal();
     void cleanNodes_internal();
-    Node* getNodeFromBlob_internal(const string* nodeSerialized);
+    std::shared_ptr<Node> getNodeFromBlob_internal(const string* nodeSerialized);
     void applyKeys_internal(uint32_t appliedKeys);
-    void notifyNode_internal(Node* node, node_vector* nodesToReport);
+    void notifyNode_internal(std::shared_ptr<Node> node, sharedNode_vector* nodesToReport);
     bool loadNodes_internal();
     uint64_t getNodeCount_internal();
     NodeCounter getCounterOfRootNodes_internal();
-    void updateCounter_internal(Node &n, Node *oldParent);
-    bool setrootnode_internal(Node* node);
+    void updateCounter_internal(std::shared_ptr<Node> n, std::shared_ptr<Node> oldParent);
+    bool setrootnode_internal(std::shared_ptr<Node> node);
     FingerprintPosition insertFingerprint_internal(Node* node);
-    void removeFingerprint_internal(Node* node);
+    void removeFingerprint_internal(Node* node, bool unloadNode);
     void saveNodeInDb_internal(Node *node);
     void dumpNodes_internal();
     void addChild_internal(NodeHandle parent, NodeHandle child, Node *node);
@@ -429,6 +616,8 @@ private:
     void setRootNodeVault_internal(NodeHandle h);
     void setRootNodeRubbish_internal(NodeHandle h);
     void initCompleted_internal();
+    void insertNodeCacheLRU_internal(std::shared_ptr<Node> node);
+    void unLoadNodeFromCacheLRU();
 };
 
 } // namespace
