@@ -27,26 +27,9 @@
 #include <algorithm>
 
 #if !defined(__MINGW32__) && !defined(__ANDROID__) && (!defined(__GNUC__) || (__GNUC__*100+__GNUC_MINOR__) >= 503)
-
-#define HAVE_FILESYSTEM
-
-#if (__cplusplus >= 201700L)
-    #if __has_include(<filesystem>)
-        #include <filesystem>
-        namespace fs = std::filesystem;
-    #else
-        #include <experimental/filesystem>
-        namespace fs = std::experimental::filesystem;
-    #endif
-#else
-#ifdef WIN32
+    #define HAVE_FILESYSTEM
     #include <filesystem>
-    namespace fs = std::experimental::filesystem;
-#else
-    #include <experimental/filesystem>
-    namespace fs = std::experimental::filesystem;
-#endif
-#endif
+    namespace fs = std::filesystem;
 #endif
 
 namespace mega {
@@ -119,11 +102,11 @@ string ACState::quoted_word::getQuoted()
 
 bool ACState::extractflag(const string& flag)
 {
-    for (auto i = words.begin(); i != words.end(); ++i)
+    for (auto it = words.begin(); it != words.end(); ++it)
     {
-        if (i->s == flag && !i->q.quoted)
+        if (it->s == flag && !it->q.quoted)
         {
-            words.erase(i);
+            words.erase(it);
             return true;
         }
     }
@@ -132,16 +115,16 @@ bool ACState::extractflag(const string& flag)
 
 bool ACState::extractflagparam(const string& flag, string& param)
 {
-    for (auto i = words.begin(); i != words.end(); ++i)
+    for (auto it = words.begin(); it != words.end(); ++it)
     {
-        if (i->s == flag)
+        if (it->s == flag)
         {
-            auto j = i;
+            auto j = it;
             ++j;
             if (j != words.end())
             {
                 param = j->s;
-                words.erase(i, ++j);
+                words.erase(it, ++j);
                 return true;
             }
         }
@@ -494,15 +477,15 @@ void Either::Add(ExecFn f, ACN n)
 bool Either::addCompletions(ACState& s)
 {
     bool stop = true;
-    int n = s.i;
-    int best_s_i = s.i;
+    unsigned n = s.i;
+    unsigned best_s_i = s.i;
     for (auto& p : eithers)
     {
-        s.i = n;
+        s.i = static_cast<unsigned>(n);
         if (!p->addCompletions(s))
         {
             stop = false;
-            best_s_i = std::max<int>(s.i, best_s_i);
+            best_s_i = std::max<unsigned>(s.i, best_s_i);
         }
     }
     s.i = best_s_i;
@@ -536,7 +519,7 @@ std::ostream& Either::describe(std::ostream& s) const
         std::ostringstream s2;
         for (int i = 0; i < int(eithers.size() * 2) - 1; ++i)
         {
-            (i & 1 ? s2 << "|" : s2 << *eithers[i / 2]);
+            (i & 1 ? s2 << "|" : s2 << *eithers[static_cast<size_t>(i / 2)]);
         }
         std::string str = s2.str();
         if (str.find(' ') == std::string::npos)
@@ -551,8 +534,9 @@ std::ostream& Either::describe(std::ostream& s) const
     return s;
 }
 
-WholeNumber::WholeNumber(size_t def_val)
-    : defaultvalue(def_val)
+WholeNumber::WholeNumber(const std::string& description, size_t defaultValue)
+  : defaultvalue(defaultValue)
+  , description(description)
 {
 }
 
@@ -567,7 +551,7 @@ bool WholeNumber::addCompletions(ACState& s)
     {
         for (char c : s.word().s)
         {
-            if (!isdigit(c))
+            if (!is_digit(static_cast<unsigned>(c)))
             {
                 return true;
             }
@@ -584,7 +568,7 @@ bool WholeNumber::match(ACState& s) const
     {
         for (char c : s.word().s)
         {
-            if (!isdigit(c))
+            if (!is_digit(static_cast<unsigned>(c)))
             {
                 return false;
             }
@@ -598,7 +582,7 @@ bool WholeNumber::match(ACState& s) const
 
 std::ostream& WholeNumber::describe(std::ostream& s) const
 {
-    return s << "N";
+    return s << description;
 }
 
 
@@ -691,7 +675,7 @@ MegaFS::MegaFS(bool files, bool folders, MegaClient* c, ::mega::NodeHandle* curD
 {
 }
 
-Node* addShareRootCompletions(ACState& s, MegaClient* client, string& pathprefix)
+shared_ptr<Node> addShareRootCompletions(ACState& s, MegaClient* client, string& pathprefix)
 {
     const string& path = s.word().s;
     string::size_type t = path.find_first_of(":/");
@@ -710,7 +694,7 @@ Node* addShareRootCompletions(ACState& s, MegaClient* client, string& pathprefix
                 string::size_type pos = path.find_first_of("/", t + 1);
                 for (handle h : u.second.sharing)
                 {
-                    if (Node* n = client->nodebyhandle(h))
+                    if (shared_ptr<Node> n = client->nodebyhandle(h))
                     {
                         if (pos == string::npos)
                         {
@@ -735,7 +719,7 @@ bool MegaFS::addCompletions(ACState& s)
     {
         if (client && cwd)
         {
-            Node* n = NULL;
+            shared_ptr<Node> n;
             std::string pathprefix;
             if (!s.word().s.empty() && s.word().s[0] == '/')
             {
@@ -790,8 +774,8 @@ bool MegaFS::addCompletions(ACState& s)
                 }
                 else
                 {
-                    Node* nodematch = NULL;
-                    for (Node* subnode : client->getChildren(n))
+                    shared_ptr<Node> nodematch = NULL;
+                    for (auto& subnode : client->getChildren(n.get()))
                     {
                         if (subnode->type == FOLDERNODE && subnode->displayname() == folderName)
                         {
@@ -813,7 +797,7 @@ bool MegaFS::addCompletions(ACState& s)
                 // iterate specified folder
                 if (n)
                 {
-                    for (Node* subnode : client->getChildren(n))
+                    for (auto& subnode : client->getChildren(n.get()))
                     {
                         if ((reportFolders && subnode->type == FOLDERNODE) ||
                             (reportFiles && subnode->type == FILENODE))
@@ -1065,13 +1049,14 @@ ACState prepACState(const std::string line, size_t insertPos, bool unixStyle)
     do
     {
         linepos = identifyNextWord(line, linepos.second);
-        std::string word = line.substr(linepos.first, linepos.second - linepos.first);
+        std::string word = line.substr(static_cast<size_t>(linepos.first),
+                                       static_cast<size_t>(linepos.second - linepos.first));
         last = linepos.first == linepos.second;
         bool cursorInWord = linepos.first <= int(insertPos) && int(insertPos) <= linepos.second;
         if (cursorInWord)
         {
             last = true;
-            word.erase(insertPos - linepos.first, std::string::npos);
+            word.erase(insertPos - static_cast<size_t>(linepos.first), std::string::npos);
             linepos.second = int(insertPos);  // keep everything to the right of the cursor
         }
         if (!acs.words.empty() && linepos.first == acs.wordPos.back().second)
@@ -1201,7 +1186,7 @@ unsigned utf8GlyphCount(const string &str)
     int c, i, ix, q;
     for (q = 0, i = 0, ix = int(str.length()); i < ix; i++, q++)
     {
-        c = (unsigned char)str[i];
+        c = (unsigned char)str[static_cast<size_t>(i)];
 
         if (c >= 0 && c <= 127) i += 0;
         else if ((c & 0xE0) == 0xC0) i += 1;
@@ -1209,13 +1194,13 @@ unsigned utf8GlyphCount(const string &str)
         else if ((c & 0xF8) == 0xF0) i += 3;
         else q++; // invalid utf8 - leave lots of space
     }
-    return q;
+    return static_cast<unsigned>(q);
 }
 
 const string& CompletionState::unixColumnEntry(int row, int col, int rows)
 {
     static string emptyString;
-    size_t index = unixListCount + col * rows + row;
+    size_t index = unixListCount + static_cast<size_t>(col * rows + row);
     return index < completions.size() ? completions[index].s : emptyString;
 }
 
@@ -1243,15 +1228,17 @@ void applyCompletion(CompletionState& s, bool forwards, unsigned consoleWidth, C
         if (!s.unixStyle)
         {
             int index = ((!forwards && s.lastAppliedIndex == -1) ? -1 : (s.lastAppliedIndex + (forwards ? 1 : -1))) + (int)s.completions.size();
-            index = static_cast<int>(index % s.completions.size());
+            index = static_cast<int>(static_cast<unsigned>(index) % s.completions.size());
 
             // restore quotes if it had them already
-            auto& c = s.completions[index];
+            auto& c = s.completions[static_cast<size_t>(index)];
             std::string w = c.s;
             s.originalWord.q.applyQuotes(w);
             w += (s.completions.size() == 1 && !c.couldExtend) ? " " : "";
-            s.line.replace(s.wordPos.first, s.wordPos.second - s.wordPos.first, w);
-            s.wordPos.second = int(w.size() + s.wordPos.first);
+            s.line.replace(static_cast<size_t>(s.wordPos.first),
+                           static_cast<size_t>(s.wordPos.second - s.wordPos.first),
+                           w);
+            s.wordPos.second = int(w.size() + static_cast<unsigned>(s.wordPos.first));
             s.lastAppliedIndex = index;
 
             if (s.completions.size()==1)
@@ -1281,8 +1268,10 @@ void applyCompletion(CompletionState& s, bool forwards, unsigned consoleWidth, C
                 }
                 s.originalWord.q.applyQuotes(exactChars);
                 exactChars += (s.completions.size() == 1 && !s.completions[0].couldExtend) ? " " : "";
-                s.line.replace(s.wordPos.first, s.wordPos.second - s.wordPos.first, exactChars);
-                s.wordPos.second = int(exactChars.size() + s.wordPos.first);
+                s.line.replace(static_cast<size_t>(s.wordPos.first),
+                               static_cast<size_t>(s.wordPos.second - s.wordPos.first),
+                               exactChars);
+                s.wordPos.second = int(exactChars.size() + static_cast<unsigned>(s.wordPos.first));
                 s.firstPressDone = true;
                 s.unixListCount = 0;
                 if (s.completions.size() == 1)
@@ -1296,7 +1285,8 @@ void applyCompletion(CompletionState& s, bool forwards, unsigned consoleWidth, C
                 unsigned rows = 1, cols = 0, sumwidth = 0;
                 for (unsigned c = 0; ;)
                 {
-                    unsigned width = s.calcUnixColumnWidthInGlyphs(c, rows);
+                    unsigned width =
+                        s.calcUnixColumnWidthInGlyphs(static_cast<int>(c), static_cast<int>(rows));
                     if (width == 0)
                     {
                         cols = c;
@@ -1331,18 +1321,22 @@ void applyCompletion(CompletionState& s, bool forwards, unsigned consoleWidth, C
                     }
                 }
 
-                rows = std::max<int>(rows, 1);
-                cols = std::max<int>(cols, 1);
+                rows = std::max<unsigned>(rows, 1);
+                cols = std::max<unsigned>(cols, 1);
                 for (unsigned c = 0; c < cols; ++c)
                 {
-                    textOut.columnwidths.push_back(s.calcUnixColumnWidthInGlyphs(c, rows) + (c == 0 ? 6 : 3));
+                    textOut.columnwidths.push_back(static_cast<int>(
+                        s.calcUnixColumnWidthInGlyphs(static_cast<int>(c), static_cast<int>(rows)) +
+                        (c == 0 ? 6 : 3)));
                 }
                 for (unsigned r = 0; r < rows; ++r)
                 {
                     textOut.stringgrid.push_back(vector<string>());
                     for (unsigned c = 0; c < cols; ++c)
                     {
-                        const string& entry = s.unixColumnEntry(r, c, rows);
+                        const string& entry = s.unixColumnEntry(static_cast<int>(r),
+                                                                static_cast<int>(c),
+                                                                static_cast<int>(rows));
                         if (!entry.empty())
                         {
                             textOut.stringgrid[r].push_back((c == 0 ? "   " : "") + entry);
@@ -1425,9 +1419,14 @@ ACN repeat(ACN n)
 }
 
 
-ACN wholenumber(size_t defaultvalue)
+ACN wholenumber(const std::string& description, size_t defaultValue)
 {
-    return std::make_shared<WholeNumber>(defaultvalue);
+    return std::make_shared<WholeNumber>(description, defaultValue);
+}
+
+ACN wholenumber(size_t defaultValue)
+{
+    return wholenumber("N", defaultValue);
 }
 
 ACN localFSPath(const std::string descriptionPrefix)
